/*
 * file ControllableResource.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM
 * 
 * (c) Copyright IBM Corporation 2004, 2008.  All Rights Reserved. 
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or  
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp. 
 */
package javax.wvcm;

import javax.wvcm.PropertyNameList.PropertyName;
import javax.wvcm.Workspace.MergeFlag;
import javax.wvcm.WvcmException.ReasonCode;

/**
 * A proxy for a controllable resource.
 * 
 * A controllable resource is always the member of
 * a special type of controllable folder called a "workspace".
 * A controllable resource is either a version-controllable resource,
 * a baseline-controllable resource, or both.
 * <p>
 * A version-controlled resource is a resource that is under control
 * of a Checkout/Checkin protocol.
 * The controlled state (content, controlled properties, and bindings to version-controlled members)
 * of a version-controlled resource cannot be modified by a client
 * unless that version-controlled resource is checked out.
 * <p>
 * When a version-controlled resource is checked in,
 * a new version of that resource is created.
 * The controlled state of the version-controlled resource is captured immutably in the new version.
 * <p>
 * The persistent state for a version-controlled resource may be
 * stored on the client, on the server, or both.
 * Storing the state on the server allows it to be shared
 * by multiple clients.  Storing the state on the client
 * provides access to that state even when the client is
 * disconnected from the server, and often provides more
 * efficient access to that state.
 * Note that the choice of what part of the state of a
 * version-controlled resource is stored on the client
 * is completely up to the implementation.  In particular,
 * although most implementations will cache some state on the client,
 * an implementation may chose to store no local state.
 * 
 * @since 1.0
 */
public interface ControllableResource extends Resource {

    // Basic Methods

    /**
     * Get the workspace provider of this resource.
     * 
     * @return the {@link WorkspaceProvider} for this Resource.
     */
    public WorkspaceProvider workspaceProvider();

    /**
     * Create a new persistent resource at the location identified by this ControllableResource.
     * The request will fail if a resource already exists at that location.
     * 
     * <p>
     * Postconditions:
     * <li>(initialize-resource): A new resource of the specified type
     *   exists at the location of this Resource.
     *   
     * @param feedback Specifies optional feedback to the caller.
     * @return A proxy for this new resource, whose properties are specified by feedback.
     * @throws WvcmException ReasonCode:
     * <li>{@link ReasonCode#RESOURCE_ALREADY_EXISTS_AT_LOCATION}:
     *  A resource MUST NOT exist at the location of this Resource.
     * <li>{@link ReasonCode#CANNOT_CREATE_AT_THIS_LOCATION}:
     *  The location of this Resource MUST identify a valid location to create this Resource.
     *  For a workspace, a client can determine locations where workspaces can be created
     *  from the {@link Resource#WORKSPACE_FOLDER_LIST} property.
     */
    public ControllableResource doCreateResource(Feedback feedback) throws WvcmException;

    /**
     * Place this resource under version-control.
     * When under version-control, the resource must be checked out to be modified
     * and checked in to commit the modifications as a new version
     * in the version history of the resource.
     * <p>
     * Postconditions:
     * <li>(put-under-version-control): If this ControllableResource identified
     *  a resource that was not under version-control at the time of the request,
     *  the request MUST have created a new version history
     *  and MUST have created a new version resource in that version history.
     *  The {@link #IS_VERSION_CONTROLLED} property of this ControllableResource MUST be true.
     *  The {@link #IS_CHECKED_OUT} property of this ControllableResource MUST be false.
     *  The resource MUST have a {@link #CHECKED_IN} property that identifies the new version.
     *  The content and controlled properties of the new version MUST be the same as that of the resource.
     *  Note that an implementation can choose to locate the version history
     *  and version resources anywhere that it wishes.  In particular, it could locate them
     *  on the same host and server as the version-controlled resource,
     *  on a different virtual host maintained by the same server,
     *  on the same host maintained by a different server, 
     *  or on a different host maintained by a different server.
     * <li>(must-not-change-existing-checked-in-out): If this ControllableResource
     *  identified a resource already under version-control at the time of the request,
     *  the request MUST NOT change the {@link #CHECKED_IN}
     *  or {@link #CHECKED_OUT} property of that version-controlled resource.
     * <li>(new-version-history): If the request created a new version history,
     *  the request MUST have allocated a new server-defined location for that version history
     *  that MUST NOT have previously identified any other resource,
     *  and MUST NOT ever identify a resource other than this version history.
     *  <li>(update-stream): If the {@link Workspace#STREAM} property of the workspace
     *  of this controllable resource is non-null, the {@link Stream#VERSION_LIST} of the stream
     *  is updated to contain the new version resource.
     *  <li>(auto-checkout-parent): If the parent of this ControllableResource
     *  was a checked-in version-controlled folder and this ControllableResource
     *  is not already under version-control, the request MAY have checked out
     *  the parent folder.
     *  
     * @param feedback Specifies optional feedback to the caller.
     * @return A new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException ReasonCode:
     * <li>{@link ReasonCode#FORBIDDEN}
     *  If the {@link #IS_VERSION_CONTROLLABLE} property of this resource is false, the request MUST fail.
     *  
     */
    public ControllableResource doVersionControl(Feedback feedback) throws WvcmException;

    // Basic Properties

    /** 
     * The workspace of which this ControllableResource is a member.
     * By definition, the workspace property of a workspace
     * identifies itself, and the workspace property of any other
     * type of resource must identify the same workspace as its parent.
     */
    public static final PropertyName<Workspace> WORKSPACE =
        new PropertyName<Workspace>("workspace"); //$NON-NLS-1$

    /**
     * Get the {@link #WORKSPACE} property.
     * 
     * @return the {@link #WORKSPACE} property.
     * @throws WvcmException if this ControllableResource was not created with
     * {@link #WORKSPACE} as a wanted property.
     */
    public Workspace getWorkspace() throws WvcmException;

    /** true if the resource is version-controlled, false otherwise. */
    public static final PropertyName<Boolean> IS_VERSION_CONTROLLED =
        new PropertyName<Boolean>("is-version-controlled"); //$NON-NLS-1$

    /**
     * Get the {@link #IS_VERSION_CONTROLLED} property.
     * 
     * @return the {@link #IS_VERSION_CONTROLLED} property.
     * @throws WvcmException if this ControllableResource was not created with
     * {@link #IS_VERSION_CONTROLLED} as a wanted property.
     */
    public boolean getIsVersionControlled() throws WvcmException;

    /** true if the resource is checked out, false otherwise. */
    public static final PropertyName<Boolean> IS_CHECKED_OUT =
        new PropertyName<Boolean>("is-checked-out"); //$NON-NLS-1$

    /**
     * Get the {@link #IS_CHECKED_OUT} property.
     * 
     * @return the {@link #IS_CHECKED_OUT} property.
     * @throws WvcmException if this ControllableResource was not created with
     * {@link #IS_CHECKED_OUT} as a wanted property.
     */
    public boolean getIsCheckedOut() throws WvcmException;


    /**
     * Create a new version-controlled resource at the location identified by the proxy.
     * The resource is associated with an existing version history,
     * and if possible, is initialized with a specified version from that version history.
     * If the specified version must be restored at a specific location,
     * the new version-controlled resource is created at that specific location
     * instead of the location of this ControllableResource.
     * The request will fail if a resource already exists at that location.
     * <p>
     * Postconditions:
     * <li>(new-version-controlled-resource): A new version-controlled resource is created
     *  for the version history of the specified version.  If possible, the initial
     *  {@link ControllableResource#CHECKED_IN} property of the new version-controlled
     *  resource is set to be the specified version; if not, the version is implementation dependent.
     *  Otherwise, same semantics as the {@link Workspace#doUpdate}.
     * <li>(new-version-controlled-folder): Same semantics as the update of a
     *  folder target in {@link Workspace#doUpdate}.
     *  <li>(auto-checkout-parent): If the parent of this ControllableResource
     *  was a checked-in version-controlled folder, the request MAY automatically check out
     *  the parent folder.
     *  
     * @param version the version used to initialize the version-controlled
     * resource.
     * @param feedback Specifies optional feedback to the caller.
     * @return A proxy for this new resource, whose properties are specified by feedback.
     * @throws WvcmException ReasonCode:
     * <li>{@link ReasonCode#BAD_ARGUMENT_TYPE}:
     * The resource identified by the <code>version</code> argument must be a version.
     * <li>{@link ReasonCode#CANNOT_CREATE_AT_THIS_LOCATION}:
     *  The location for the new version-controlled resource is not a valid location for a version-controlled resource.
     * <li>{@link ReasonCode#RESOURCE_ALREADY_EXISTS_AT_LOCATION}:
     *  The location for the new version-controlled resource MUST NOT identify an existing resource.
     * <li>{@link ReasonCode#ONE_CONTROLLED_RESOURCE_PER_HISTORY_PER_WORKSPACE}:
     *  The workspace of this ControllableResource MUST NOT have a member
     *  whose {@link #VERSION_HISTORY} property identifies the version history
     *  of the version specified in the request. 
     */
    public ControllableResource doCreateVersionControlledResource(Version version, Feedback feedback)
    throws WvcmException;

    /** Boolean flags for the doCheckin method */
    public static class CheckinFlag
    {
        private CheckinFlag(String flagName) { _flagName = flagName; }

        /**
         * Returns a string representation of this flag suitable for diagnostics.
         */
        @Override
        public String toString() { return _flagName; }

        /**
         * A string representation of this flag.
         */
        private final String _flagName;

        /** 
         * Indicates whether to fork even if forking is discouraged.
         */
        public static final CheckinFlag FORK_OK = new CheckinFlag("fork-ok"); //$NON-NLS-1$

        /** 
         * Indicates whether to checkout the resource again immediately after checking it in.
         */
        public static final CheckinFlag KEEP_CHECKED_OUT = new CheckinFlag("keep-checked-out"); //$NON-NLS-1$

        /**
         * Indicates whether a checkin will be performed if the persistent state
         * of the new version would be identical to that of the previous
         * version. If this flag is not set, a checkin operation on an unchanged
         * controllable resource will uncheckout that resource.
         */
        public static final CheckinFlag CHECKIN_IDENTICAL = new CheckinFlag("checkin-identical"); //$NON-NLS-1$
    }


    /** Boolean flags for the doCheckout method */
    public static class CheckoutFlag 
    {
        private CheckoutFlag(String flagName) { _flagName = flagName; }

        /**
         * Returns a string representation of this flag suitable for diagnostics.
         */
        @Override
        public String toString() { return _flagName; }

        /**
         * The string form of this flag.
         */
        private final String _flagName;

        /** 
         * Indicates whether to fork even if forking is discouraged.
         */
        public static final CheckoutFlag FORK_OK = new CheckoutFlag("fork-ok"); //$NON-NLS-1$

        /**
         * Indicates whether to fail the checkout if there already is a reserved checkout
         * from the currently selected version in the checkout activity.
         */
        @SuppressWarnings("hiding") //$NON-NLS-1$
        public static final CheckoutFlag RESERVED = new CheckoutFlag("reserved"); //$NON-NLS-1$
    }


    /**
     * Changes the state of the resource identified by this ControllableResource
     * to be checked-in, and synchronizes the persistent state
     * of the resource to be that of the persistent state
     * on the client.  The current content and controlled properties of the resource are captured
     * in a new version resource at a server-defined location.
     * A doCheckin request can be applied to a checked-out configuration
     * to produce a new baseline whose {@link Baseline#VERSION_LIST} captures the
     * current state of the version-controlled members of the {@link Configuration#ROOT_FOLDER} 
     * of the configuration.
     * <p>
     * Postconditions:
     * <li>(uncheckout-identical): If the controlled state of this resource is the same as its CHECKED_OUT version,
     * and if the PREDECESSOR_LIST only contains the CHECKED_OUT version,
     * and if the flag {@link CheckinFlag#CHECKIN_IDENTICAL} has not been passed in,
     * an uncheckout on this resource will be automatically performed.
     * <li>(complete-merge): If the {@link #MERGE_LIST} or the {@link #AUTO_MERGE_LIST} are non-empty,
     * the versions within them are first moved to {@link #PREDECESSOR_LIST}.
     * <li>(create-version): The request MUST have created a new version in the version history
     *  of the {@link #CHECKED_OUT} version.  The request MUST have allocated a distinct new location
     *  for the new version, and that location MUST NOT ever identify any resource other than that version.
     * <li>(initialize-version-content-and-properties): The content and controlled properties
     *  of the new version MUST be copied from the checked-out resource.
     *  The {@link Version#VERSION_NAME} of the new version MUST be set to a server-defined value
     *  distinct from all other {@link Version#VERSION_NAME} values of other versions in the same
     *  version history.
     *  The {@link #COMMENT} and {@link #PREDECESSOR_LIST} of the version-controlled resource
     *  MUST have been set to <code>null</code> after it was copied to the new version.
     * <li>(checked-in): If keepCheckedOut is not specified in the request,
     *  the {@link #CHECKED_OUT} property of the version-controlled resource MUST have been removed
     *  and a {@link #CHECKED_IN} property that identifies the new version MUST have been added.
     * <li>(keep-checked-out): If keepCheckedOut is specified in the request,
     *  the {@link #CHECKED_OUT} property of the checked-out resource MUST have been updated
     *  to identify the new version.
     * <li>(add-to-history): The new version resource MUST have been added as a new bound member
     *  of the version history of the {@link #CHECKED_OUT} version.
     * <li>(initialize-activity): The {@link Version#ACTIVITY} of the new version
     *  MUST have been initialized to be the same as the {@link #ACTIVITY} of the checked-out resource. 
     *  The {@link #ACTIVITY} of the version-controlled resource MUST have been set to NULL
     *  after it was copied to the new version.
     * <li>(initialize-folder-version-bindings): If this ControllableResource identifies
     *  a version-controlled folder, then for each version-controlled bound member of this
     *  ControllableResource, the new folder version MUST contain a bound member which
     *  is the version history of that version-controlled bound member, and whose binding
     *  name is the binding name of the version-controlled bound member.
     * <li>(create-baseline-version-list): If this ControllableResource identifies a configuration,
     *  the {@link Baseline#VERSION_LIST} of the new baseline identifies a
     *  list of the {@link #CHECKED_IN} versions of each version-controlled member
     *  of the {@link Configuration#ROOT_FOLDER} of the configuration at the time of the request.
     *  <li>(update-stream): The {@link Stream#VERSION_LIST} of the {@link Workspace#STREAM}
     *  is updated to contain the new version resource.
     *  If the {@link Stream#VERSION_LIST} had previously contained a different version from the version history of the new version,
     *  that different version is removed.
     *  
     *  @param flags Array of boolean flags (may be null):
     * <li>{@link CheckinFlag#KEEP_CHECKED_OUT} indicates whether to checkout the
     * resource again immediately after checking it in.
     * If <code>true</code>, the {@link #ACTIVITY}
     * and {@link #RESERVED} properties should not be changed by the checkout.
     * <li>{@link CheckinFlag#FORK_OK} indicates whether to fork even if forking is discouraged.
     * <li>{@link CheckinFlag#CHECKIN_IDENTICAL} will force a checkin even if the controlled
     * state is the same as its CHECKED_IN version. Otherwise, an uncheckout will be performed if
     * the controlled state of this resource is the same as its CHECKED_IN version,
     * and if the PREDECESSOR_LIST only contains the CHECKED_OUT version.
     * @param feedback Specifies optional feedback to the caller.
     * @return A new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException ReasonCode:
     * <li>{@link ReasonCode#MUST_BE_CHECKED_OUT}:
     *  This ControllableResource MUST identify a resource whose {@link #IS_CHECKED_OUT} property is true.
     * <li>{@link ReasonCode#VERSION_HISTORY_MUST_BE_A_TREE}:
     *  The versions identified by the {@link #PREDECESSOR_LIST}, {@link #MERGE_LIST} 
     *  and {@link #AUTO_MERGE_LIST} of the checked-out resource MUST be descendants of 
     *  the root version of the version history for the {@link #CHECKED_OUT} version.
     * <li>{@link ReasonCode#CANNOT_FORK}:
     *  A doCheckin request MUST fail if it would cause a version
     *  whose {@link Version#CHECKIN_FORK} is {@link Version.Fork#FORBIDDEN} to 
     *  appear in the {@link Version#PREDECESSOR_LIST} of more than one version.
     * <li>{@link ReasonCode#CANNOT_CHECKIN_FORK_DISCOURAGED}:
     *  A doCheckin request MUST fail if it would cause a version
     *  whose {@link Version#CHECKIN_FORK} is {@link Version.Fork#DISCOURAGED} to 
     *  appear in the {@link Version#PREDECESSOR_LIST} of more than one version,
     *  unless forkOk is specified in the request.
     * <li>{@link ReasonCode#CANNOT_CREATE_BRANCH_IN_ACTIVITY}:
     *  Any version which is in the version history of the checked-out resource
     *  and whose {@link Version#ACTIVITY} identifies the {@link #ACTIVITY}
     *  of the checked-out resource MUST be an ancestor of the
     *  checked-out resource.  
     * <li>@{@link ReasonCode#CANNOT_CHECKIN_TO_RESERVED_ACTIVITY}
     *  If in another workspace there is a version-controlled resource
     *  whose {@link #CHECKED_OUT} property identifies a version in the version history
     *  of this resource, then the {@link #RESERVED} property of that version-controlled
     *  resource MUST be true. 
     * <li>{@link ReasonCode#NO_CHECKED_OUT_BASELINE_CONTROLLED_FOLDER_MEMBERS}:
     *  If this ControllableResource
     *  identifies a configuration, all version-controlled members
     *  of the {@link Configuration#ROOT_FOLDER} of the configuration MUST be checked-in.
     *  <li>{@link ReasonCode#CANNOT_CREATE_BRANCH_IN_STREAM}:
     *  If the {@link Stream#VERSION_LIST} of the {@link Workspace#STREAM} of this ControllableResource
     *  contains a version from the version history of this controllable resource,
     *  then the {@link Stream#VERSION_LIST}
     *  MUST contain one of the versions in the {@link #PREDECESSOR_LIST} of this controllable resource.
     *  
     */
    public ControllableResource doCheckin(CheckinFlag[] flags, Feedback feedback)
    throws WvcmException;

    /**
     * Checkout this ControllableResource so that its controlled state can be modified.
     * <p>
     * Postconditions:
     * <li>(is-checked-out): The resource identified by this ControllableResource
     *  MUST have an {@link #IS_CHECKED_OUT} property whose value is true.
     * <li>(has-checked-out-version): The checked-out resource MUST have a {@link #CHECKED_OUT} property
     *  that identifies the {@link #CHECKED_IN} version preceding the checkout.
     *  The version-controlled resource MUST NOT have a {@link #CHECKED_IN} property.
     * <li>(initialize-predecessor-list): The {@link #PREDECESSOR_LIST} property of the checked-out resource
     *  MUST be initialized to be the {@link #CHECKED_OUT} version.
     * <li>(initialize-activity): The {@link #ACTIVITY} of the checked-out resource is set as follows:
     *  - If the resource is a configuration, then the activity is the {@link Workspace#STREAM} of
     *    the workspace (possibly null).
     *  - Otherwise, if the resource is not a configuration, the {@link Workspace#CURRENT_ACTIVITY}
     *    of the workspace is used.  If the {@link Workspace#CURRENT_ACTIVITY} is <code>null</code>,
     *    the server MAY automatically set the {@link Workspace#CURRENT_ACTIVITY} to be a newly created activity.
     * <li>(initialize-reserved): If reserved was specified in the request,
     *  then the {@link #RESERVED} property of the checked-out resource MUST be true.
     * @param flags Array of boolean flags (may be null):
     * @param feedback Specifies optional feedback to the caller.
     * @return A new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException ReasonCode:
     * <li>{@link ReasonCode#MUST_BE_CHECKED_IN}:
     *  The {@link #IS_CHECKED_OUT} property of this ControllableResource MUST be false.
     * <li>{@link ReasonCode#CANNOT_CHECKOUT_FORKING_IS_FORBIDDEN}:
     *  If the {@link Version#CHECKOUT_FORK} property
     *  of the version being checked out is {@link Version.Fork#FORBIDDEN}, the request MUST fail
     *  if the checked-out version has a non-empty list of successors.
     * <li>{@link ReasonCode#CANNOT_CHECKOUT_FORKING_IS_DISCOURAGED}:
     *  If the {@link Version#CHECKOUT_FORK} property
     *  of the version being checked out is {@link Version.Fork#DISCOURAGED}, the request MUST fail
     *  if the checked-out version has a non-empty list of successors unless forkOk is specified 
     *  in the request.
     * <li>{@link ReasonCode#CANNOT_CHECKOUT_MULTI_CHECKOUT_IS_DISCOURAGED}:
     *  If the {@link Version#CHECKOUT_FORK} property
     *  of the version being checked out is {@link Version.Fork#DISCOURAGED}, the request MUST fail
     *  if a checked-out resource identifies that version in its {@link #CHECKED_OUT} property
     *  unless forkOk is specified in the request.
     * <li>{@link ReasonCode#CANNOT_CHECKOUT_MULTI_CHECKOUT_IS_FORBIDDEN}:
     *  If the {@link Version#CHECKOUT_FORK} property
     *  of the version being checked out is {@link Version.Fork#FORBIDDEN}, the request MUST fail
     *  if a checked-out resource identifies that version in its {@link #CHECKED_OUT} property.
     * <li>{@link ReasonCode#ONE_CHECKOUT_PER_ACTIVITY_PER_HISTORY}:
     *  If there is a request activity set,
     *  and if Reserved is specified, an existing reserved checkout from a version
     *  of that version history MUST NOT select an activity in that activity set.
     * <li>{@link ReasonCode#CANNOT_CREATE_BRANCH_IN_ACTIVITY}:
     *  If there is a request activity set, and if Reserved is specified,
     *  the selected version MUST be a descendant of all other versions of that version history
     *  that select that activity. 
     *  <li>{@link ReasonCode#CANNOT_CREATE_BRANCH_IN_STREAM}:
     *  If the {@link Workspace#STREAM} property of the workspace of this controllable resource
     *  is non-null, one of the following MUST be true:
     *  <br> - the {@link Stream#VERSION_LIST} of the stream contains no version from the version history
     *  of the {@link #CHECKED_IN} version.
     *  <br> - the {@link Stream#VERSION_LIST} of the stream contains the {@link #CHECKED_IN} version,
     *  <br> - the {@link CheckoutFlag#RESERVED} was not specified in the flags argument.
     *  <li>{@link ReasonCode#BAD_ARGUMENT_TYPE}:
     *  If the controllable resource is a configuration, and there is a request activity set,
     *  the <code>activity</code> argument MUST identify a stream; otherwise, the <code>activity</code>
     *  MUST identify an activity.
     */
    public ControllableResource doCheckout(
            CheckoutFlag[] flags,
            Feedback feedback)
    throws WvcmException;

    /**
     * Cancels the checkout of a version-controlled resource,
     * and restores its controlled state to the state of its {@link #CHECKED_OUT} version.
     * <p>
     * Postconditions:
     * <li>(cancel-checked-out): The value of the {@link #CHECKED_IN} property is that of the
     *  {@link #CHECKED_OUT} property prior to the request, and the {@link #CHECKED_OUT} property has been removed.
     * <li>(restore-content): The controlled state of the version-controlled resource
     *  corresponds to that of its {@link #CHECKED_IN} version.
     *  
     * @param feedback Specifies optional feedback to the caller.
     * @return A new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException ReasonCode:
     * <li>{@link ReasonCode#MUST_BE_CHECKED_OUT}:
     *  This ControllableResource
     *  MUST identify a version-controlled resource with a {@link #CHECKED_OUT} property.
     */
    public ControllableResource doUncheckout(Feedback feedback) throws WvcmException;

    /**
     * Add the specified label to the checked-in version of this controllable resource.
     * @see Version#doAddLabel
     */
    public ControllableResource doAddLabel(String label, Feedback feedback) throws WvcmException;

    /**
     * Remove the specified lable from the {@link #CHECKED_IN} version of this controllable resource.
     * @see Version#doRemoveLabel
     */
    public ControllableResource doRemoveLabel(String label, Feedback feedback) throws WvcmException;

    /**
     * Set the specified label on the checked-in version of this controllable resource.
     * @see Version#doSetLabel
     */
    public ControllableResource doSetLabel(String label, Feedback feedback) throws WvcmException;

    /**
     * Apply {@link Resource#doReadProperties}
     * to the version selected by <code>label</code> from the version history
     * of this ControllableResource.
     * 
     * @param label the label identifying the version to be selected 
     * @param feedback the list of properties that will be available on
     * the returned version.
     * @return a proxy for the selected version, with the wanted properties.
     * @throws WvcmException ReasonCode:
     * <li>{@link ReasonCode#MUST_BE_CHECKED_IN}:
     *  The {@link #CHECKED_IN} property of this ControllableResource MUST identify a version.
     * <li>{@link ReasonCode#NOT_FOUND}:
     *  The label was not found.
     */
    public Version doReadLabelledVersionProperties(
            String label,
            Feedback feedback)
    throws WvcmException;

    // Versioning Properties

    /**
     * A boolean value indicating whether this resource can be placed under version-control.
     * @see #doVersionControl doVersionControl
     */
    public static final PropertyName<Boolean> IS_VERSION_CONTROLLABLE =
        new PropertyName<Boolean>("is-version-controllable"); //$NON-NLS-1$

    /**
     * Get the {@link #IS_VERSION_CONTROLLABLE} property.
     * 
     * @return the {@link #IS_VERSION_CONTROLLABLE} property.
     * @throws WvcmException if this ControllableResource was not created with
     * {@link #IS_VERSION_CONTROLLABLE} as a wanted property.
     */
    public boolean getIsVersionControllable() throws WvcmException;

    /**
     * The version history of a version-controlled resource.
     * This is the version history of the {@link #CHECKED_IN} or {@link #CHECKED_OUT} version.
     */
    public static final PropertyName<VersionHistory> VERSION_HISTORY =
        new PropertyName<VersionHistory>("version-history"); //$NON-NLS-1$

    /**
     * Get the {@link #VERSION_HISTORY} property.
     * 
     * @return the {@link #VERSION_HISTORY} property.
     * @throws WvcmException if this ControllableResource was not created with
     * {@link #VERSION_HISTORY} as a wanted property.
     */
    public VersionHistory getVersionHistory() throws WvcmException;

    /**
     * The configuration to which this resource belongs.
     */
    public static final PropertyName<Configuration> CONFIGURATION =
        new PropertyName<Configuration>("configuration"); //$NON-NLS-1$

    /**
     * Get the {@link #CONFIGURATION} property.
     * 
     * @return the {@link #CONFIGURATION} property.
     * @throws WvcmException if this property was not set and
     * this ControllableResource was not created with
     * {@link #CONFIGURATION} as a wanted property.
     */
    public Configuration getConfiguration()
    throws WvcmException;

    /**
     * The checked-in {@link Version} of this version-controlled resource.
     * If this ControllableResource is checked-in,
     * this is the version whose content and controlled properties
     * are identical to those of this ControllableResource;
     * otherwise, this is <code>null</code>.
     */
    public static final PropertyName<Version> CHECKED_IN =
        new PropertyName<Version>("checked-in"); //$NON-NLS-1$

    /**
     * Get the {@link #CHECKED_IN} property.
     * 
     * @return the {@link #CHECKED_IN} property.
     * @throws WvcmException if this ControllableResource was not created with
     * {@link #CHECKED_IN} as a wanted property.
     */
    public Version getCheckedIn() throws WvcmException;

    /**
     * The checked-out version of a version-controlled resource.
     * if this ControllableResource is checked-out,
     * this is the checked-in version at the time
     * the checkout was performed; otherwise, this is <code>null</code>.
     */
    public static final PropertyName<Version> CHECKED_OUT =
        new PropertyName<Version>("checked-out"); //$NON-NLS-1$

    /**
     * Get the {@link #CHECKED_OUT} property.
     * 
     * @return the {@link #CHECKED_OUT} property.
     * @throws WvcmException if this ControllableResource was not created with
     * {@link #CHECKED_OUT} as a wanted property.
     */
    public Version getCheckedOut() throws WvcmException;

    /** 
     * If this ControllableResource is checked-out,
     * this is a list of the versions that
     * will become the predecessors of the version created
     * when this ControllableResource is checked in;
     * otherwise, this is <code>null</code>.
     * @see #getPredecessorList
     * @see #setPredecessorList
     */
    public static final PropertyName<ResourceList<Version>> PREDECESSOR_LIST =
        new PropertyName<ResourceList<Version>>("predecessor-list"); //$NON-NLS-1$

    /**
     * Get the {@link #PREDECESSOR_LIST} property.
     * 
     * @return the {@link #PREDECESSOR_LIST} property.
     * @throws WvcmException if this property was not set and
     * this ControllableResource was not created with
     * {@link #PREDECESSOR_LIST} as a wanted property.
     * @see #setPredecessorList
     */
    public ResourceList<Version> getPredecessorList() throws WvcmException;

    /**
     * Set the {@link #PREDECESSOR_LIST} property.
     * 
     * @param versionList a list of versions that will
     * be the predecessors of the version created when this ControllableResource
     * is checked in.
     * @see #getPredecessorList
     */
    public void setPredecessorList(ResourceList<Version> versionList);

    /**
     * If the controllable resource is checked-out, this identifies the activity
     * that will become the {@link Version#ACTIVITY} of the version created
     * when the controllable resource is checked in; otherwise, this is <code>null</code>.
     * If the controllable resource is not checked-out, an attempt to modify this
     * property MUST fail.
     * If the controllable resource is a Configuration, and the {@link #ACTIVITY} is not <code>null</code>,
     * the activity must be a {@link Stream}.
     * @see #setActivity
     * @see #getActivity
     */
    public static final PropertyName<Activity> ACTIVITY =
        new PropertyName<Activity>("activity"); //$NON-NLS-1$

    /**
     * Get the {@link #ACTIVITY} property.
     * 
     * @return the {@link #ACTIVITY} property.
     * @throws WvcmException if this property was not set and
     * this ControllableResource was not created with
     * {@link #ACTIVITY} as a wanted property.
     * @see #setActivity
     */
    public Activity getActivity() throws WvcmException;

    /**
     * Set the {@link #ACTIVITY} property.
     * 
     * @param activity a list of {@link Activity} objects that identify
     * the new value for the {@link #ACTIVITY} property.
     * @see #getActivity
     */
    public void setActivity(Activity activity);

    /**
     * A Boolean that indicates whether the checked-out version
     * of this checked-out version-controlled resource can
     * concurrently be checked-out to the same activity in 
     * another version-controlled resource.
     * <p>In particular, this indicates whether the {@link #ACTIVITY} 
     * of another checked-out resource associated with the
     * version history of this version-controlled resource
     * can have be the activity that is also identified by the
     * {@link #ACTIVITY} property of this checked-out resource.
     * Since an activity must form a single line of descent through
     * a given version history, a <code>doCheckin</code> may fail
     * when a checkout is not reserved, because another version-controlled
     * resource may have checked-in a parallel version into the
     * activity.  In this case, the user will first have
     * to merge into this checked-out resource the latest
     * version selected by that activity from that version history,
     * and then modify the {@link #PREDECESSOR_LIST} of this checked-out resource
     * to identify that version. 
     * @see #setReserved
     * @see #getReserved
     */
    public static final PropertyName<Boolean> RESERVED =
        new PropertyName<Boolean>("reserved"); //$NON-NLS-1$

    /**
     * Get the {@link #RESERVED} property.
     * 
     * @return the {@link #RESERVED} property.
     * @throws WvcmException if this property was not set and
     * this ControllableResource was not created with
     * {@link #RESERVED} as a wanted property.
     * @see #setReserved
     */
    public boolean getReserved() throws WvcmException;

    /**
     * Set the {@link #RESERVED} property.
     * 
     * @param val the new value for the {@link #RESERVED} property of this ControllableResource.
     * @see #getReserved
     */
    public void setReserved(boolean val);

    /**
     * A list of {@link Version} objects that identify the versions whose
     * content and controlled properties must be merged by the client into this
     * checked-out ControllableResource.
     * <p>
     * After a merge operation, the merge list of each merge target contains a
     * list of merge source versions, indicating the versions whose content or
     * controlled properties are yet to be merged. For each entry in the merge
     * list, it is the client's responsibility to update the controlled state of
     * the checked-out merge target so that it reflects the logical merge of the
     * merge source into the current controlled state of the merge target. The
     * client indicates that it has completed the update of the merge target by
     * deleting the merge source from the merge list of the checked-out merge
     * target, and adding it to the predecessor list.
     * <p>
     * As an error check for a client forgetting to complete a merge, the server
     * MUST fail an attempt to <code>doCheckin()</code> a version-controlled
     * resource with a non-empty merge list.
     * 
     * @see #setMergeList
     * @see #getMergeList
     */
    public static final PropertyName<ResourceList<Version>> MERGE_LIST =
        new PropertyName<ResourceList<Version>>("merge-list"); //$NON-NLS-1$

    /**
     * Get the {@link #MERGE_LIST} property.
     * 
     * @return the {@link #MERGE_LIST} property.
     * @throws WvcmException if this property was not set and
     * this ControllableResource was not created with
     * {@link #MERGE_LIST} as a wanted property.
     * @see #setMergeList
     */
    public ResourceList<Version> getMergeList() throws WvcmException;

    /**
     * Set the {@link #MERGE_LIST} property.
     * 
     * @param versionList a list of {@link Version} objects that will be
     * the value of the {@link #MERGE_LIST} property.
     * @see #getMergeList
     */
    public void setMergeList(ResourceList<Version> versionList);

    /**
     * A list of {@link Version} objects that identify each version whose
     * controlled state has been merged by the server into this checked-out
     * ControllableResource.
     * <p>
     * When a server has the ability to automatically update the controlled
     * state of the merge target to reflect the logical merge of the merge
     * source, it may do so unless {@link MergeFlag#NO_AUTO_MERGE} is specified
     * in the {@link Workspace#doMerge} request.
     * <p>
     * After a merge operation, the auto-merge list of each merge target
     * contains a list of merge source versions whose controlled state was
     * automatically merged into that merge target. Automatically merged source
     * versions appear only in the auto-merge list, and not the merge list.
     * <p>
     * The client is responsible for deleting entries from the
     * {@link #AUTO_MERGE_LIST} and adding them to the {@link #PREDECESSOR_LIST}
     * after the user has verified that the automatic merge was performed
     * correctly by the server.
     * 
     * @see #setAutoMergeList
     * @see #getAutoMergeList
     */
    public static final PropertyName<ResourceList<Version>> AUTO_MERGE_LIST =
        new PropertyName<ResourceList<Version>>("auto-merge-list"); //$NON-NLS-1$

    /**
     * Get the {@link #AUTO_MERGE_LIST} property.
     * 
     * @return the {@link #AUTO_MERGE_LIST} property.
     * @throws WvcmException if this property was not set and
     * this ControllableResource was not created with
     * {@link #AUTO_MERGE_LIST} as a wanted property.
     * @see #setAutoMergeList
     */
    public ResourceList<Version> getAutoMergeList() throws WvcmException;

    /**
     * Set the {@link #AUTO_MERGE_LIST} property.
     * 
     * @param versionList a list of {@link Version} objects that identify
     * the new value for the {@link #AUTO_MERGE_LIST} property.
     * @see #getAutoMergeList
     */
    public void setAutoMergeList(ResourceList<Version> versionList);

}
